from typing import Any


class Prompt:
    def __init__(self, template_str: str) -> None:
        self.template = template_str

    def format(self, **kwargs: Any) -> str:
        return self.template.format(**kwargs)



from typing import List, Dict
from tinylang.images import Image
from tinylang.llms.base import BaseLLM
from tinylang.memory import ConversationMemory
from tinylang.memory.base import BaseMemory
from tinylang.messages import SystemMessage


class Chain:
    def __init__(
        self,
        llm: BaseLLM,
        memory: BaseMemory | None = None,
        prompt: str | SystemMessage | None = None,
        tools: List | None = None,
    ) -> None:
        self.llm = llm
        self.memory = memory or ConversationMemory()
        self.llm.memory = self.memory
        self.tools = tools

        if isinstance(prompt, SystemMessage):
            self.memory.add_message(prompt)

        elif isinstance(prompt, str):
            self.memory.add_system_message(prompt)

    def run(
        self,
        prompt: str,
        raw_response: bool = False,
        image: Image | None = None,
        **kwargs: Dict,
    ) -> str:
        return self.llm.chat(prompt, raw_response, image, **kwargs)



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict

from tinylang.images import Image
from .utils import from_json


@dataclass  # type: ignore
class BaseMessage(ABC):
    prefix: str
    text: str
    image: Image | None
    image_dir: os.PathLike

    def __init__(
        self,
        text: str,
        prefix: str,
        image: Image | None = None,
    ) -> None:
        self.prefix = prefix
        self.text = text
        self.image = image

    def to_json(self) -> Dict:
        result = {"role": self.prefix}

        content = [{"type": "text", "text": self.text}]

        if self.image is not None:
            content.append({"type": "image_url", "image_url": self.image.to_url()})

        result["content"] = content

        return result

    @staticmethod
    @abstractmethod
    def from_json(json: Dict) -> "BaseMessage":
        return from_json(json)  # type: ignore

    def copy(self) -> "BaseMessage":
        return self.from_json(self.to_json())

    def __str__(self) -> str:
        return str(self.to_json())

    def __repr__(self) -> str:
        return str(self.to_json())



import base64
import io
import os
from pathlib import Path
from typing import Dict

import numpy as np
from PIL import Image as PillowImage


class Image:
    def __init__(
        self,
        image_bytes: bytes,
        format: str = "JPEG",
        image_dir: os.PathLike = Path(__name__).parent.parent,
    ) -> None:
        self.image_bytes = image_bytes
        self.format = format.upper()
        self.image_dir = image_dir

    def to_numpy(self) -> np.ndarray:
        with io.BytesIO(self.image_bytes) as image_io:
            with PillowImage.open(image_io) as pil_image:
                return np.array(pil_image.convert("RGB"))

    def to_b64(self) -> str:
        return base64.b64encode(self.image_bytes).decode("utf-8")

    @classmethod
    def from_pillow(cls, pil_image: PillowImage, format: str = "JPEG") -> "Image":
        with io.BytesIO() as output:
            pil_image.save(output, format=format)
            return cls(output.getvalue(), format)

    @classmethod
    def from_opencv(cls, opencv_image: np.ndarray, format: str = "JPEG") -> "Image":
        pil_image = PillowImage.fromarray(opencv_image)
        return cls.from_pillow(pil_image, format)

    @classmethod
    def from_numpy(cls, np_image_data: np.ndarray, format: str = "JPEG") -> "Image":
        pil_image = PillowImage.fromarray(np_image_data)
        return cls.from_pillow(pil_image, format)

    @classmethod
    def from_bytes(cls, image_bytes: bytes) -> "Image":
        with io.BytesIO(image_bytes) as image_io:
            pil_image = PillowImage.open(image_io)
            format = pil_image.format
            if format not in ["JPEG", "PNG", "WEBP", "GIF"]:
                raise ValueError("Unsupported image format")
            return cls(image_bytes, format)

    def to_bytes(self) -> bytes:
        return self.image_bytes

    def to_dict(self) -> Dict:
        base64_encoded = self.to_b64()
        return {
            "image_data": base64_encoded,
            "dtype": str(self.to_numpy().dtype),
            "shape": self.to_numpy().shape,
            "format": self.format,
        }

    def to_url(self) -> str:
        return f"data:image/{self.format.lower()};base64,{self.to_b64()}"

    @classmethod
    def from_dict(cls, data: Dict) -> "Image":
        raw_bytes = base64.b64decode(data["image_data"])
        return cls(raw_bytes, data.get("format", "JPEG"))

    def __str__(self) -> str:
        return f"Image object with {self.format} format, size: {self.to_numpy().shape}"



from typing import Dict

from tinylang.images import Image
from tinylang.messages.base import BaseMessage


class UserMessage(BaseMessage):
    def __init__(
        self,
        content: str,
        prefix: str = "user",
        image: Image | None = None,
    ) -> None:
        super().__init__(content, prefix, image)

    @staticmethod
    def from_json(json: Dict[str, str]) -> "UserMessage":
        return UserMessage(
            content=json["content"],
            image=json.get("image"),  # type: ignore
            prefix=json.get("role", "user"),
        )



from collections import deque
from dataclasses import dataclass
from typing import Dict, List

from ..messages import AssistantMessage, SystemMessage, UserMessage
from ..messages.base import BaseMessage
from .base import BaseMemory
from tinylang.images import Image


@dataclass
class ConversationMemory(BaseMemory):
    last_k: int | None

    def __init__(
        self,
        messages: List[BaseMessage] | None = None,
        last_k: int | None = None,
    ) -> None:
        self._last_k_validation(last_k)
        super().__init__(messages)

        self.last_k = last_k
        self._last_k_messages = None if last_k is None else last_k * 2
        # * 2 to keep assistant and user messages

        # TODO: fix bug if using a system message

        # use a deque to keep the last k messages
        self.messages = deque(self.messages, maxlen=self._last_k_messages)

    @staticmethod
    def _last_k_validation(last_k: int | None) -> int | None:
        if last_k is None:
            return last_k

        if last_k < 1:
            raise ValueError("last_k must be positive")

        return last_k

    def add_message(self, message: BaseMessage) -> None:
        self.messages.append(message)

    def add_user_message(
        self, message: str, prefix: str = "user", image: Image | None = None
    ) -> None:
        self.add_message(UserMessage(message, prefix, image))

    def add_system_message(self, message: str) -> None:
        self.add_message(SystemMessage(message))

    def add_assistant_message(self, message: str) -> None:
        self.add_message(AssistantMessage(message))

    def to_json(self) -> List[Dict]:
        return [message.to_json() for message in self.messages]

    def format_messages(self) -> List[Dict]:
        temp = [message.to_json() for message in self.messages]

        if self._last_k_messages is None:
            return temp

        return temp[-self._last_k_messages :]



from functools import wraps
import inspect
from typing import Any, Callable, get_type_hints

from pydantic import create_model


# TODO: check reddit post
def function(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    # Extract annotations and defaults
    annotations = get_type_hints(f)
    defaults = dict(inspect.signature(f).parameters)

    fields_dict = {}
    for name, type_ in annotations.items():
        default_value = (
            defaults[name].default
            if defaults[name].default != inspect.Parameter.empty
            else None
        )
        fields_dict[name] = (type_, default_value)

    input_model = create_model(f"InputModel_{f.__name__}", **fields_dict)

    # Generate JSON schema
    schema = input_model.model_json_schema()

    # Attach metadata to the wrapper function
    wrapper.schema = dict(name=f.__name__, description=f.__doc__, parameters=schema)

    return wrapper



import os
from typing import Any, Dict, Generator

from openai import OpenAI as OpenAIClient

from tinylang.images import Image
from tinylang.llms.base import BaseLLM
from tinylang.memory.base import BaseMemory
from tinylang.messages import UserMessage


class OpenAI(BaseLLM):
    def __init__(
        self,
        openai_api_key: str,
        openai_organization: str,
        model: str,
        memory: BaseMemory | None = None,
        **kwargs: Dict,
    ) -> None:
        super().__init__(memory)
        self.openai_api_key = openai_api_key
        self.openai_organization = openai_organization
        self.model = model
        self.kwargs = (
            kwargs  # kwargs passed into the openai.ChatCompletion.create() method
        )

        self.client = OpenAIClient(
            api_key=self.openai_api_key, organization=self.openai_organization
        )

    def load_model(self, model_path: os.PathLike) -> bool:
        """
        Loads a local model from a path.
        """
        return True  # since we're using the API

    def chat(
        self,
        prompt: str,
        raw_response: bool = False,
        image: Image | None = None,
        **kwargs: Dict,
    ) -> str:
        # TODO: move this elsewhere
        # be consistent with the prefix
        prefix = "user"
        for message in self.memory.messages:
            if isinstance(message, UserMessage):
                prefix = message.prefix
                break

        self.memory.add_user_message(message=prompt, prefix=prefix, image=image)

        completion_kwargs = self.kwargs.copy()
        completion_kwargs.update(kwargs)

        api_response = self.client.chat.completions.create(
            model=self.model,
            messages=self.memory.format_messages(),
            **completion_kwargs,
        )  # type: ignore

        chat_response: str = api_response.model_dump()["choices"][0]["message"][
            "content"
        ]
        self.memory.add_assistant_message(chat_response)

        if raw_response:
            return api_response  # type: ignore

        return chat_response

    def stream_chat(
        self, prompt: str, raw_response: bool = False, **kwargs: Dict
    ) -> Generator[Dict[str, Any], None, None]:
        self.memory.add_user_message(prompt)

        completion_kwargs = self.kwargs.copy()
        completion_kwargs.update(kwargs)

        api_response = self.client.chat.completions.create(
            model=self.model,
            messages=self.memory.format_messages(),
            stream=True,
            **completion_kwargs,
        )  # type: ignore

        aggregated_content = ""
        for chunk in api_response:
            content = chunk.choices[0].delta.content or ""

            # if there's no content
            if not content:
                continue

            aggregated_content += content

            yield chunk if raw_response else content

        self.memory.add_assistant_message(aggregated_content)
